import { computed, MaybeRefOrGetter, readonly, ref, shallowRef, toRef, toValue, watch } from 'vue'
import { wrap as wrapWorker } from 'comlink'
import type { ExposedInterface } from './vips.worker'
import { ensureError } from '@/util'
import { useEventListener } from '@vueuse/core'

const _vipsWorkerError = ref<Error | null>(null)
const supportsImageOperations = computed(() => _vipsWorkerError.value === null)

function onWorkerError(event: ErrorEvent) {
  _vipsWorkerError.value = ensureError(event.error ?? event.message)
}

export const vipsWorkerError = readonly(_vipsWorkerError)

export function useImageOperations(image: MaybeRefOrGetter<Blob | null>) {
  const vipsWorker = new Worker(new URL('./vips.worker', import.meta.url), { type: 'module' })
  useEventListener(vipsWorker, 'error', onWorkerError)
  const vips = wrapWorker<ExposedInterface>(vipsWorker)
  const width = ref<number | undefined>()
  const height = ref<number | undefined>()
  const isModified = ref(false)
  const dimensions = computed(() =>
    typeof width.value === 'undefined' || typeof height.value === 'undefined'
      ? undefined
      : { width: width.value, height: height.value },
  )
  const isLoadingNewImage = ref(false)
  const isRunningOperation = ref(false)
  const isWorking = computed(() => isLoadingNewImage.value || isRunningOperation.value)
  const renderedImage = shallowRef<Blob | null>(toValue(image))

  watch(
    toRef(image),
    async (image) => {
      if (!image) return
      renderedImage.value = image
      if (!supportsImageOperations.value) return
      isLoadingNewImage.value = true
      try {
        await vips.setImage(image)
        await updateDimensions()
      } finally {
        isLoadingNewImage.value = false
      }
    },
    { immediate: true },
  )

  async function updateDimensions() {
    if (!supportsImageOperations.value) return
    const dimensions = await vips.getDimensions()
    width.value = dimensions.width
    height.value = dimensions.height
  }

  function wrap<T>(operation: (...args: T[]) => Promise<void>) {
    return async (...args: T[]) => {
      if (!supportsImageOperations.value) return

      isRunningOperation.value = true
      try {
        await operation(...args)
        const newImage = await vips.renderImage()
        isModified.value = true
        renderedImage.value = newImage
        await updateDimensions()
      } catch (e) {
        console.error('libvips operation error', e)
        throw e
      } finally {
        isRunningOperation.value = false
      }
    }
  }

  return {
    isWorking,
    supportsImageOperations: supportsImageOperations,
    isModified: readonly(isModified),
    image: readonly(renderedImage),
    width: readonly(width),
    height: readonly(height),
    dimensions,
    crop: wrap(vips.crop),
    flip: wrap(vips.flip),
    rotate: wrap(vips.rotate),
  }
}
