import { defineStore } from 'pinia'
import { ensureError, makeInsecureRandomId } from '@/util'
import { computed, MaybeRefOrGetter, reactive, ref, toValue, watch } from 'vue'
import { File as TankFile, Media, MediaSourceCreateData, Show } from '@/types'
import { ImportOptions, useFilesStore } from '@/stores/files'
import { useMediaStore } from '@/stores/media'
import { useMediaSourceStore } from '@/stores/media-sources'
import { APIResponseError } from '@rokoli/bnb/drf'

type FailReason = 'notAnAudioFile' | 'cancelledByUser' | 'unableToUpload' | undefined

type MediaSourceStatus =
  // attached to the queue (initial state)
  | { key: 'queued' }
  // rejected by the queue processor
  | { key: 'rejectedLocally'; reason?: FailReason }
  // rejected by the server
  | { key: 'rejectedRemotely'; reason?: FailReason }
  // job was cancelled (most likely by the user)
  | { key: 'cancelled'; reason?: FailReason }
  // failed to process by the queue processor (retry possible)
  | { key: 'failed'; reason?: FailReason }
  // failed to process by the queue processor (retry not possible)
  | { key: 'failedPermanently'; reason?: FailReason }
  // currently being processed by the queue processor
  | {
      key: 'processing'
      progress?: number
      phase?: string
    }

interface FileData {
  file: File
}

interface FileImportData {
  importUrl: string
}

interface URLData {
  url: string
}

export type MediaResolver = (options?: { useCached?: boolean } | undefined) => Promise<Media>

export interface MediaSourceJob {
  id: string
  name: string
  contextKey: string | null
  status: MediaSourceStatus
  data: FileData | FileImportData | URLData
  controller: AbortController
  error: Error | null
  show: Show
  getMedia: MediaResolver
  tankFile: TankFile | null
  lastProcessor: MediaSourceJobProcessor | null
}

type MediaSourceJobProcessor = (job: MediaSourceJob, next: () => void) => Promise<void>

const ensureFileIsAudio: MediaSourceJobProcessor = async function (job, next) {
  if ('file' in job.data && !job.data.file.type.startsWith('audio/')) {
    job.status = { key: 'rejectedLocally', reason: 'notAnAudioFile' }
    return
  }
  next()
}

const importFile: MediaSourceJobProcessor = async (job: MediaSourceJob, next: () => void) => {
  if (!('file' in job.data || 'importUrl' in job.data)) return next()

  const fileStore = useFilesStore()

  job.status = { key: 'processing' }

  if (job.tankFile) {
    // A tank file already exists for this file,
    // which means that this is a retry.
    await fileStore.remove(job.tankFile.id)
    job.tankFile = null
    job.error = null
  }

  // While we’re uploading, watch the progress and update the job status.
  const progressWatcher = watch(
    () => (job.tankFile ? fileStore.importProgressMap.get(job.tankFile.id) ?? null : null),
    (progress) => {
      if (!progress) return
      job.status = {
        key: 'processing',
        phase: `file:${progress.step}`,
        progress: progress.progress,
      }
    },
  )

  // These options are shared for uploads and URL imports.
  const importOptions: ImportOptions = {
    requestInit: { signal: job.controller.signal },
    onDone() {
      job.status = { key: 'processing' }
    },
    onCreate(file) {
      job.tankFile = file
    },
  }

  try {
    if ('file' in job.data) {
      await fileStore.uploadFile(job.data.file, job.show, importOptions)
    } else if ('importUrl' in job.data) {
      await fileStore.importFileURL(job.data.importUrl, job.show, importOptions)
    }
  } catch (err) {
    if (job.controller.signal.aborted) throw err
    job.error = ensureError(err)
    job.status = { key: 'failed', reason: 'unableToUpload' }
    return
  } finally {
    progressWatcher.stop()
  }

  if (job.tankFile) {
    // refresh tank file so that it has the correct duration.
    job.tankFile = await fileStore.retrieve((job.tankFile as TankFile).id, { useCached: false })
  }

  next()
}

async function makeMediaSourceFromJob(job: MediaSourceJob): Promise<MediaSourceCreateData> {
  const mediaId = (await job.getMedia()).id
  if (job.tankFile) {
    return { mediaId, duration: job.tankFile.duration, fileId: job.tankFile.id }
  } else if ('url' in job.data) {
    return { mediaId, duration: null, uri: job.data.url }
  } else {
    throw new TypeError('Invalid job type. Cannot convert to media source.')
  }
}

const attachMediaSource: MediaSourceJobProcessor = async function attachMediaSource(job, next) {
  const mediaSourceStore = useMediaSourceStore()
  job.status = { key: 'processing', phase: 'media:save' }
  try {
    await mediaSourceStore.create(await makeMediaSourceFromJob(job))
  } catch (e) {
    if (e instanceof APIResponseError) {
      job.error = e
      const { status } = e.response
      if (status >= 400 && status < 500) {
        job.status = { key: 'failed' }
        return
      }
    }

    throw e
  }
  next()
}

const mediaSourceJobProcessors: MediaSourceJobProcessor[] = [
  ensureFileIsAudio,
  importFile,
  attachMediaSource,
]

const useMediaSourceManagerStore = defineStore('mediaSourceQueue', () => {
  const jobQueueMap = ref(new Map<string, MediaSourceJob>())
  const jobQueue = computed(() => Array.from(jobQueueMap.value.values()))

  function removeJob(id: string) {
    jobQueueMap.value.delete(id)
  }

  function enqueue(
    baseJobData: Pick<MediaSourceJob, 'name' | 'data' | 'contextKey' | 'show' | 'getMedia'>,
  ) {
    const id = makeInsecureRandomId()
    const job = reactive({
      ...baseJobData,
      id,
      status: { key: 'queued' } as const,
      controller: new AbortController(),
      error: null,
      lastProcessor: null,
      tankFile: null,
    })
    jobQueueMap.value.set(id, job)
    _process(id)
    return job
  }

  async function dismiss(id: string) {
    const job = jobQueueMap.value.get(id)
    removeJob(id)

    if (job) {
      const filesStore = useFilesStore()
      const { controller, tankFile } = job
      if (!controller.signal.aborted) {
        job.controller.abort()
      }
      if (tankFile) {
        await filesStore.remove(tankFile.id)
      }
    }
  }

  async function startNextProcessor(job: MediaSourceJob) {
    const processor =
      job.lastProcessor === null
        ? mediaSourceJobProcessors[0]
        : mediaSourceJobProcessors[mediaSourceJobProcessors.indexOf(job.lastProcessor) + 1]

    if (!processor) {
      // If we haven’t found a processor, there’s nothing left to do for this job record.
      removeJob(job.id)
      return
    }

    try {
      job.error = null
      await processor(job, () => {
        job.lastProcessor = processor
        process(job.id)
      })
    } catch (err) {
      if (job.controller.signal.aborted) {
        job.status = { key: 'cancelled', reason: job.controller.signal.reason }
      } else {
        job.error = ensureError(err)
        job.status = { key: 'failedPermanently' }
      }
    }
  }

  function canProcess(job: MediaSourceJob) {
    const validStates: MediaSourceJob['status']['key'][] = [
      'queued',
      'processing',
      'failed',
      'cancelled',
    ]
    return validStates.includes(job.status.key)
  }

  function _process(id: MediaSourceJob['id']) {
    const job = jobQueueMap.value.get(id)
    if (!job) return

    // Don’t process jobs that shouldn’t be processed.
    // This is the case for jobs that have been rejected or failed permanently.
    if (!canProcess(job)) return

    if (job.controller.signal.aborted) {
      // Job processing was aborted before, but because this function
      // will only be called in case of a retry, we can safely create a new
      // job controller that is used for this re-run.
      job.controller = new AbortController()
    }

    // We don’t care for the async return state of this function
    // as all of its state is handled through properties on
    // the job object.
    void startNextProcessor(job)
  }

  function process(...ids: MediaSourceJob['id'][]) {
    for (const id of ids) {
      setTimeout(_process.bind(null, id), 0)
    }
  }

  return {
    jobQueueMap,
    jobQueue,
    process,
    dismiss,
    enqueue,
  }
})

export function useMediaFactory(
  show: MaybeRefOrGetter<Show>,
  media: MaybeRefOrGetter<Media | null | undefined>,
  assignMedia: (media: Media) => Promise<unknown>,
): MediaResolver {
  const mediaStore = useMediaStore()
  return async function getMedia(options) {
    const _media: Media | null = toValue(media) ?? null

    if (_media !== null) {
      const refreshedMedia = await mediaStore.retrieve(_media.id, options)
      if (refreshedMedia) return refreshedMedia
    }

    const showId = toValue(show).id
    const newMedia = await mediaStore.create({ showId })
    try {
      await assignMedia(newMedia)
    } catch (err) {
      // The media instance didn’t exist before so we should delete it
      // if assignment fails.
      await mediaStore.remove(newMedia.id)
      // We can’t actually handle the assignment error here,
      // so we need to re-throw the error.
      throw err
    }
    return newMedia
  }
}

export function useBasicMediaSourceController() {
  const store = useMediaSourceManagerStore()

  function dismiss(id: string) {
    void store.dismiss(id)
  }

  function cancel(id: string, reason: FailReason = 'cancelledByUser') {
    const job = store.jobQueueMap.get(id)
    if (!job) return
    job.controller.abort(reason)
  }

  function retry(id: string) {
    store.process(id)
  }

  return { dismiss, cancel, retry }
}

export function useMediaSourceController(options: {
  show: MaybeRefOrGetter<Show>
  getMedia: MediaResolver
  contextKey?: MaybeRefOrGetter<string | null | undefined>
}) {
  const store = useMediaSourceManagerStore()
  const basicOperations = useBasicMediaSourceController()

  function enqueue(name: string, entry: MediaSourceJob['data']) {
    return store.enqueue({
      name,
      data: entry,
      contextKey: toValue(options.contextKey) ?? null,
      show: toValue(options.show),
      getMedia: options.getMedia,
    })
  }

  function addFiles(...files: (FileList | File | File[] | null | undefined)[]) {
    const _files = files
      .filter((f) => f !== null && f !== undefined)
      .flatMap((f) => (f instanceof FileList ? Array.from(f) : f)) as File[]
    for (const file of _files) {
      enqueue(file.name, { file })
    }
  }

  function addImports(...importUrls: string[]) {
    for (const importUrl of importUrls) {
      const name = importUrl.split('/').at(-1) ?? ''
      enqueue(name, { importUrl })
    }
  }

  function addUrls(...data: (string | { url: string; name: string })[]) {
    for (const item of data) {
      if (typeof item === 'string') {
        const name = item.split('/').at(-1) ?? ''
        enqueue(name, { url: item })
      } else {
        enqueue(item.name, { url: item.url })
      }
    }
  }

  return {
    ...basicOperations,
    addFiles,
    addImports,
    addUrls,
  }
}

export function useMediaSourceQueue(contextKey: MaybeRefOrGetter<string | null | undefined>) {
  const store = useMediaSourceManagerStore()

  return computed(() => {
    const _contextKey = toValue(contextKey)
    return _contextKey
      ? store.jobQueue.filter((entry) => entry.contextKey === _contextKey)
      : Array.from(store.jobQueue)
  })
}

export function useJobHasState(
  job: MaybeRefOrGetter<MediaSourceJob>,
  states: MediaSourceJob['status']['key'][],
) {
  return computed(() => states.includes(toValue(job).status.key))
}
