import { defineStore } from 'pinia'
import { createSteeringURL } from '@/api'
import {
  APICreate,
  APIListPaginated,
  APIRemove,
  APIRetrieve,
  APIUpdate,
  createExtendableAPI,
} from '@rokoli/bnb/drf'
import { MediaSource as _MediaSource, MediaSourceCreateData, MediaSourceUpdateData } from '@/types'
import { steeringAuthInit } from '@/stores/auth'
import { useMediaStore } from '@/stores/media'
import { aggregateWithIdsParameter, APIReorder } from '@/util/api'

export function determineMediaSourceType(mediaSource: _MediaSource) {
  if (mediaSource.fileId !== null) return 'file' as const

  let url
  try {
    url = new URL(mediaSource.uri.toLowerCase())
  } catch (e) {
    return 'unknown' as const
  }

  const protocol = url.protocol
  if (protocol === 'm3u:') return 'm3u' as const
  if (protocol === 'line:') return 'line' as const
  if (protocol === 'http:' || protocol === 'https:') return 'stream' as const

  return 'unknown' as const
}

export const useMediaSourceStore = defineStore('mediaSources', () => {
  const endpoint = createSteeringURL.prefix('media-sources')
  const { api, base } = createExtendableAPI<_MediaSource>(endpoint, steeringAuthInit)
  const listOperations = APIListPaginated(api)

  function refreshMediaFromEvent({ detail }: { detail: { object: _MediaSource | null } }) {
    if (!detail.object) return
    const mediaStore = useMediaStore()
    void mediaStore.retrieve(detail.object.mediaId, { useCached: false })
  }

  api.events.addEventListener('set', refreshMediaFromEvent)
  api.events.addEventListener('remove', refreshMediaFromEvent)

  return {
    ...base,
    ...listOperations,
    ...APICreate<_MediaSource, MediaSourceCreateData>(api),
    ...APIUpdate<_MediaSource, MediaSourceUpdateData, MediaSourceUpdateData>(api),
    ...APIRetrieve(api, {
      aggregate: aggregateWithIdsParameter(listOperations.listIsolated),
    }),
    ...APIReorder(api),
    ...APIRemove(api),
  }
})
