import {
  APICreate,
  APIListPaginated,
  APIRemove,
  APIRetrieve,
  APIUpdate,
  createExtendableAPI,
  useObjectFromStore,
} from '@rokoli/bnb/drf'
import { defineStore } from 'pinia'

import { createSteeringURL } from '@/api'
import { steeringAuthInit } from '@/stores/auth'
import { Media, MediaSource, MediaCreateData, MediaUpdateData } from '@/types'
import { computed, ComputedRef, MaybeRefOrGetter, toValue } from 'vue'
import { aggregateWithIdsParameter } from '@/util/api'
import { useI18n } from '@/i18n'
import { computedAsync } from '@vueuse/core'

export function calculateMediaDurationInSeconds(
  mediaSources: MediaSource[],
  skipUnknown?: true,
): number
export function calculateMediaDurationInSeconds(
  mediaSources: MediaSource[],
  skipUnknown?: false,
): number | null

/**
 * Calculates the duration of a media container.
 * May return null if the media object contains entries with unset duration.
 * @param mediaSources
 * @param skipUnknown Whether unknown length entries should be skipped
 */
export function calculateMediaDurationInSeconds(mediaSources: MediaSource[], skipUnknown = false) {
  let duration = 0
  for (const mediaSource of mediaSources) {
    // mediaSource.duration may be null/NaN if the media source references
    // a stream or other resources without an inherent duration
    if (typeof mediaSource.duration !== 'number' || isNaN(mediaSource.duration)) {
      if (skipUnknown) continue
      else return null
    }
    duration += mediaSource.duration
  }

  return duration
}

export function countUnknownDurations(mediaSources: MediaSource[]) {
  let counter = 0
  for (const mediaSource of mediaSources) {
    if (typeof mediaSource.duration !== 'number' || isNaN(mediaSource.duration)) counter += 1
  }
  return counter
}

type MediaCascadeSource = 'episode' | 'schedule' | 'show' | 'fallbackShow'

type MediaCascadeDescriptor<T extends MediaCascadeSource> = {
  mediaId: Media['id'] | null | undefined
  inheritedFrom?: T
}

export function useMediaCascade<T extends MediaCascadeSource>(
  ...cascadeDescriptors: MaybeRefOrGetter<MediaCascadeDescriptor<T>>[]
) {
  const mediaStore = useMediaStore()

  const cascade = computedAsync(
    async () => {
      const descriptors = cascadeDescriptors.map((descriptor) => toValue(descriptor))
      const isEvaluating =
        descriptors.length > 0 && descriptors.some((descriptor) => descriptor.mediaId === undefined)

      // prefetch media
      await mediaStore.retrieveMultiple(
        descriptors.filter((d) => typeof d.mediaId === 'number').map((d) => d.mediaId) as number[],
        { useCached: true },
      )

      for (const { mediaId, inheritedFrom } of descriptors) {
        if (typeof mediaId !== 'number') continue
        const media = await mediaStore.retrieve(mediaId, { useCached: true })
        if (!media) continue
        if (media.entries.length === 0) continue
        return { mediaId, isEvaluating, inheritedFrom: inheritedFrom }
      }

      return { mediaId: null, isEvaluating, inheritedFrom: null }
    },
    { mediaId: null, isEvaluating: true, inheritedFrom: null },
  )

  const { obj: media, isLoading: isLoadingMedia } = useObjectFromStore(
    () => cascade.value.mediaId,
    mediaStore,
  )
  const isEvaluatingMedia = computed(() => cascade.value.isEvaluating || isLoadingMedia.value)
  const mediaInheritedFrom = computed(() => cascade.value.inheritedFrom ?? null)
  const mediaInheritanceLabel = computed(() => {
    if (isEvaluatingMedia.value) return undefined
    if (!mediaInheritedFrom.value) return undefined
    const { t } = useI18n()
    return t(`media.inheritance.${mediaInheritedFrom.value}`)
  })

  return { media, isEvaluatingMedia, mediaInheritedFrom, mediaInheritanceLabel }
}

type MediaState =
  | 'ok'
  | 'autoExpanded'
  | 'invalid.missing'
  | 'invalid.multipleUnknownDurations'
  | 'invalid.tooLong'
  | 'notice.tooShort'

export function useMediaState(
  media: MaybeRefOrGetter<Media | null>,
  targetDurationSeconds?: MaybeRefOrGetter<number | undefined> | undefined,
): ComputedRef<{ state: MediaState; durationSeconds: number; offsetSeconds: number }> {
  return computed(() => {
    const sources = toValue(media)?.entries

    if (sources === undefined)
      return { state: 'invalid.missing', durationSeconds: 0, offsetSeconds: 0 }

    const targetDuration = toValue(targetDurationSeconds)
    const roundedTargetDuration =
      typeof targetDuration === 'number' ? Math.round(targetDuration) : null

    const unknownDurationCount = countUnknownDurations(sources)
    if (unknownDurationCount > 1) {
      // This should never happen because steering validates that we don’t have more than one
      // source of unknown duration, but better safe than sorry.
      return {
        state: 'invalid.multipleUnknownDurations',
        durationSeconds: 0,
        offsetSeconds: 0,
      }
    }

    const staticMediaDuration = Math.round(calculateMediaDurationInSeconds(sources, true))
    if (roundedTargetDuration !== null && staticMediaDuration > roundedTargetDuration) {
      return {
        state: 'invalid.tooLong',
        durationSeconds: staticMediaDuration,
        offsetSeconds: staticMediaDuration - roundedTargetDuration,
      }
    }

    // If the media contains just one record of unknown length
    // the playout will automatically expand that entry to the remaining
    // time that is needed to fill the timeslot.
    // We can therefore consider a single entry of unknown duration to fit the required time.
    if (unknownDurationCount === 1 && roundedTargetDuration !== null) {
      return {
        state: 'autoExpanded',
        durationSeconds: roundedTargetDuration,
        offsetSeconds: Math.round(roundedTargetDuration - staticMediaDuration),
      }
    }

    // There aren’t any sources that are auto-expanded, so if the static media duration is
    // too short, and we have a target duration to reach, the sources won’t fill them entirely.
    if (roundedTargetDuration !== null && staticMediaDuration < roundedTargetDuration) {
      return {
        state: 'notice.tooShort',
        durationSeconds: staticMediaDuration,
        offsetSeconds: roundedTargetDuration - staticMediaDuration,
      }
    }

    return {
      state: 'ok',
      durationSeconds: staticMediaDuration,
      offsetSeconds: 0,
    }
  })
}

export const useMediaStore = defineStore('media', () => {
  const endpoint = createSteeringURL.prefix('media')
  const { api, base } = createExtendableAPI<Media>(endpoint, steeringAuthInit)
  const listOperations = APIListPaginated(api)
  return {
    ...base,
    ...listOperations,
    ...APICreate<Media, MediaCreateData>(api),
    ...APIUpdate<Media, MediaUpdateData, MediaUpdateData>(api),
    ...APIRetrieve(api, {
      aggregate: aggregateWithIdsParameter(listOperations.listIsolated),
    }),
    ...APIRemove(api),
  }
})
