import {
  APICreate,
  APIListPaginated,
  APIRemove,
  APIRetrieve,
  APIUpdate,
  createExtendableAPI,
  ExtendableAPI,
  useObjectFromStore,
} from '@rokoli/bnb/drf'
import { defineStore } from 'pinia'
import { computed, ref, watch } from 'vue'
import { useRoute } from 'vue-router'

import { createSteeringURL } from '@/api'
import { components } from '@/steering-types'
import { steeringAuthInit, useAuthStore, useOnAuthBehaviour } from '@/stores/auth'
import { KeysFrom, Show } from '@/types'
import { aggregateWithIdsParameter } from '@/util/api'
import { useCurrentRadioSettings } from '@/stores/radio-settings'

type ReadonlyAttrs = KeysFrom<Show, 'id' | 'createdAt' | 'createdBy' | 'updatedAt' | 'updatedBy'>
type ShowCreateData = Omit<Show, ReadonlyAttrs | 'slug'> & { slug?: string }
type ShowUpdateData = ShowCreateData
type ShowPartialUpdateData = components['schemas']['PatchedShow']
export type NewShow = ShowCreateData

export function newShow(): NewShow {
  return {
    name: '',
    description: '',
    shortDescription: '',
    email: '',
    internalNote: '',
    typeId: -1,
    fundingCategoryId: -1,
    isActive: true,
    isPublic: true,
    predecessorId: null,
    defaultMediaId: null,
    cbaSeriesId: null,
    imageId: null,
    logoId: null,
    links: [],
    topicIds: [],
    categoryIds: [],
    musicFocusIds: [],
    ownerIds: [],
    hostIds: [],
    languageIds: [],
  }
}

function useSelectedShowBehaviour(api: ExtendableAPI<Show>) {
  const authStore = useAuthStore()
  const route = useRoute()

  const { retrieve } = APIRetrieve(api)

  const selectedShowId = ref<number | null>(null)

  const selectedShow = computed(() =>
    selectedShowId.value !== null ? api.itemMap.value.get(selectedShowId.value) ?? null : null,
  )

  // Ensure that we always have the currently selected show in the store
  watch(
    [selectedShowId, () => authStore.currentUser],
    ([showId, user]) => {
      if (showId && user) void retrieve(showId, { useCached: true })
    },
    { immediate: true },
  )

  useOnAuthBehaviour(() => {
    selectedShowId.value = null
  })

  // Change the selected show if the user accessed a route for a particular show.
  // This is made so the show selector and the current route don’t have mismatching shows, which may cause confusion.
  watch(
    () => route.params.showId,
    (showId) => {
      if (showId) {
        selectedShowId.value = parseInt(showId as string)
      }
    },
    { immediate: true },
  )

  return { selectedShowId, selectedShow }
}

export function useFallbackShow() {
  const showStore = useShowStore()
  const radioSettings = useCurrentRadioSettings()
  return useObjectFromStore(() => radioSettings.value?.program?.fallback?.showId ?? null, showStore)
}

export const useShowStore = defineStore('shows', () => {
  const endpoint = createSteeringURL.prefix('shows')
  const { api, base } = createExtendableAPI<Show>(endpoint, steeringAuthInit)
  const { selectedShow, selectedShowId } = useSelectedShowBehaviour(api)
  const listOperations = APIListPaginated(api)

  return {
    ...base,
    ...listOperations,
    ...APIRetrieve(api, {
      aggregate: aggregateWithIdsParameter(listOperations.listIsolated),
    }),
    ...APICreate<Show, ShowCreateData>(api),
    ...APIUpdate<Show, ShowUpdateData, ShowPartialUpdateData>(api),
    ...APIRemove(api),
    selectedShowId,
    selectedShow,
  }
})
