import {
  APIListPaginated,
  APIRemove,
  APIRetrieve,
  APIUpdate,
  createExtendableAPI,
  ExtendableAPI,
  useObjectFromStore,
} from '@rokoli/bnb/drf'
import { defineStore } from 'pinia'

import { createSteeringURL } from '@/api'
import { steeringAuthInit } from '@/stores/auth'
import { Episode, KeysFrom, TimeSlot } from '@/types'
import { aggregateWithIdsParameter } from '@/util/api'
import { useEpisodeStore } from '@/stores/episodes'
import { useFallbackShow, useShowStore } from '@/stores/shows'
import { useScheduleStore } from '@/stores/schedules'
import { MaybeRefOrGetter, toValue } from 'vue'
import { useMediaCascade } from '@/stores/media'

type ReadonlyAttrs = KeysFrom<
  TimeSlot,
  'id' | 'start' | 'end' | 'repetitionOfId' | 'scheduleId' | 'showId'
>
type TimeSlotUpdateData = Omit<TimeSlot, ReadonlyAttrs>
type TimeSlotPartialUpdateData = Partial<TimeSlotUpdateData>

export const useTimeSlotStore = defineStore('timeslots', () => {
  const endpoint = createSteeringURL.prefix('timeslots')
  const { api, base } = createExtendableAPI<TimeSlot>(endpoint, steeringAuthInit)
  const listOperations = APIListPaginated(api)

  useEpisodeUpdateCascadeBehaviour(api)

  return {
    ...base,
    ...listOperations,
    ...APIRetrieve(api, {
      aggregate: aggregateWithIdsParameter(listOperations.listIsolated),
    }),
    ...APIUpdate<TimeSlot, TimeSlotUpdateData, TimeSlotPartialUpdateData>(api),
    ...APIRemove(api),
  }
})

function useEpisodeUpdateCascadeBehaviour(api: ExtendableAPI<TimeSlot>) {
  async function fetchUpdates(episodeIds: (TimeSlot['episodeId'] | undefined)[]) {
    const episodeStore = useEpisodeStore()
    const timeslotStore = useTimeSlotStore()
    const episodeResults = await episodeStore.retrieveMultiple(
      episodeIds.filter((id) => Boolean(id)) as number[],
    )
    await timeslotStore.retrieveMultiple(
      episodeResults
        .filter((result) => result.status === 'fulfilled')
        .flatMap((result) => (result as PromiseFulfilledResult<Episode>).value.timeslotIds),
    )
  }

  api.events.addEventListener('set', (event) => {
    const oldEpisodeId = api.itemMap.value.get(event.detail.id)?.episodeId
    if (oldEpisodeId === undefined) return
    const newEpisodeId = event.detail.object.episodeId
    if (oldEpisodeId !== newEpisodeId)
      setTimeout(() => void fetchUpdates([newEpisodeId, oldEpisodeId]), 0)
  })
}

export function useTimeslotMediaCascade(timeslot: MaybeRefOrGetter<TimeSlot | null>) {
  const showStore = useShowStore()
  const scheduleStore = useScheduleStore()
  const episodeStore = useEpisodeStore()

  const { obj: episode } = useObjectFromStore(
    () => toValue(timeslot)?.episodeId ?? null,
    episodeStore,
  )
  const { obj: schedule } = useObjectFromStore(
    () => toValue(timeslot)?.scheduleId ?? null,
    scheduleStore,
  )
  const { obj: show } = useObjectFromStore(() => toValue(timeslot)?.showId ?? null, showStore)
  const { obj: fallback } = useFallbackShow()

  return useMediaCascade(
    () => ({ mediaId: episode.value?.mediaId }),
    () => ({ mediaId: schedule.value?.defaultMediaId, inheritedFrom: 'schedule' as const }),
    () => ({ mediaId: show.value?.defaultMediaId, inheritedFrom: 'show' as const }),
    () => ({ mediaId: fallback.value?.defaultMediaId, inheritedFrom: 'fallbackShow' as const }),
  )
}
