import { expect, Page, test } from '@playwright/test'
import { getFileMetadata, purgeSteeringState, purgeTankState, steeringAPI } from './util'
import { Show } from '@/types'

test.describe.configure({ mode: 'serial' })

test.beforeEach(async ({ request }) => {
  await purgeSteeringState(request, { params: { modelCategories: 'program' } })
})

async function createShow(page: Page): Promise<Show> {
  await page.goto('/')
  await page.getByTestId('nav:shows').click()
  await page.getByTestId('my-shows:add-show').click()
  await page.getByTestId('add-show-modal:show-name').fill('my series')
  await page.getByTestId('add-show-modal:show-description').fill('my series description')
  await page.getByTestId('add-show-modal:show-type').selectOption({ index: 0 })
  await page.getByTestId('add-show-modal:show-funding-category').selectOption({ index: 0 })
  await page.locator('[data-testid="add-show-modal"] button[type="submit"]').click()
  await page.waitForURL(/\/shows\/\d+/)
  const showId = (page.url().match(/\/shows\/(\d+)/) as RegExpMatchArray)[1]
  return await page.request.get(steeringAPI('shows', showId)).then((res) => res.json())
}

test('Can create new show.', async ({ page }) => {
  await createShow(page)
  await expect(page.getByTestId('page-header:lead')).toHaveText('my series')
})

test('Can delete a show', async ({ page }) => {
  const show = await createShow(page)
  await page.goto(`/shows/${show.id}/basic-data`)
  await page.getByTestId('show-deletion-flow:delete').click()
  const promptValue = await page
    .getByTestId('confirm-dialog:prompt-label')
    .locator('code')
    .innerText()
  await page.getByTestId('confirm-dialog:prompt').fill(promptValue)
  await page.getByTestId('confirm-dialog:confirm').click()

  expect((await page.request.get(steeringAPI('shows', show.id))).status()).toBe(404)
})

test.describe('Show media management', () => {
  let show: Show

  test.afterEach(async ({ request }) => {
    await purgeTankState(request)
  })

  test.beforeEach(async ({ page }) => {
    show = await createShow(page)
    await page.goto(`/shows/${show.id}/basic-data`)
  })

  test('Can upload file', async ({ page }) => {
    const fileChooserPromise = page.waitForEvent('filechooser')
    const soundfile = getFileMetadata('meeresrauschen.opus')
    await page.getByTestId('media-editor:open-file-dialog').click()
    const fileChooser = await fileChooserPromise
    await fileChooser.setFiles(soundfile.path)
    const fileTitle = await page.getByTestId('media-source-editor:file-title').innerText()
    expect(fileTitle).toBe(soundfile.name)
  })
})
